# Core imports
from __future__ import annotations
from pathlib import Path

# Internal imports
from quicfire_tools.utils import read_dat_file
from quicfire_tools.inputs import SimulationInputs
from quicfire_tools.outputs import (
    _process_grid_info,
    _get_resolution_from_coords,
    _process_fire_indexes,
    _process_compressed_bin,
    _process_gridded_bin,
    SimulationOutputs,
    OutputFile,
)

# External imports
import zarr
import numpy as np
import netCDF4 as nc

TEST_DIR = Path(__file__).parent
TEST_DATA_DIR = TEST_DIR / "data"
SAMPLE_DATA_DIR = TEST_DATA_DIR / "samples"
LINE_FIRE_DIR = SAMPLE_DATA_DIR / "LineFire"
EG_CANOPY_DIR = SAMPLE_DATA_DIR / "EglinCanopyTest"
COS_HILL_DIR = SAMPLE_DATA_DIR / "CosHill"
TMP_DIR = TEST_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)


class TestProcessGridInfo:
    def test_read_grid_info_line_fire(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        quic_nz, quic_grid, en2atmos_nz, en2atmos_grid = _process_grid_info(
            LINE_FIRE_DIR / "Output" / "grid.bin",
            line_fire.qu_simparams.ny,
            line_fire.qu_simparams.nx,
        )
        assert quic_nz == line_fire.qu_simparams.nz
        assert len(quic_grid) == line_fire.qu_simparams.nz + 2
        quic_dz = _get_resolution_from_coords(quic_grid)
        assert np.allclose(quic_dz[:-1], line_fire.qu_simparams._dz_array)

        assert en2atmos_nz == 21
        en2atmos_dz = _get_resolution_from_coords(en2atmos_grid)
        assert en2atmos_dz == line_fire.quic_fire.dz

    def test_read_grid_info_eglin_canopy(self):
        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        quic_nz, quic_grid, en2atmos_nz, en2atmos_grid = _process_grid_info(
            EG_CANOPY_DIR / "Output" / "grid.bin",
            eglin_canopy.qu_simparams.ny,
            eglin_canopy.qu_simparams.nx,
        )
        assert quic_nz == eglin_canopy.qu_simparams.nz
        assert len(quic_grid) == eglin_canopy.qu_simparams.nz + 2
        quic_dz = _get_resolution_from_coords(quic_grid)
        assert np.allclose(quic_dz[:-1], eglin_canopy.qu_simparams._dz_array)

        assert en2atmos_nz == 43
        en2atmos_dz = _get_resolution_from_coords(en2atmos_grid)
        assert en2atmos_dz == eglin_canopy.quic_fire.dz


class TestProcessedCompressedBin:
    line_fire_sim = SimulationInputs.from_directory(LINE_FIRE_DIR)
    line_fire_indexes = _process_fire_indexes(
        LINE_FIRE_DIR / "Output" / "fire_indexes.bin"
    )

    eglin_canopy_sim = SimulationInputs.from_directory(EG_CANOPY_DIR)
    eglin_canopy_indexes = _process_fire_indexes(
        EG_CANOPY_DIR / "Output" / "fire_indexes.bin"
    )

    def test_line_fire_fuel_dens(self):
        fuels_dens_start = _process_compressed_bin(
            LINE_FIRE_DIR / "Output" / "fuels-dens-00000.bin",
            (
                self.line_fire_sim.quic_fire.nz,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
            self.line_fire_indexes,
        )
        assert fuels_dens_start.shape == (
            1,
            self.line_fire_sim.quic_fire.nz,
            self.line_fire_sim.qu_simparams.ny,
            self.line_fire_sim.qu_simparams.nx,
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "fuels-dens-00000.npz")
        reference_fuel_dens = npz["var"]
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 0, 2)

        assert np.allclose(fuels_dens_start[0], reference_fuel_dens)

        # trim xy edges of the array
        fuels_dens_start = fuels_dens_start[0, 0, 1:-1, 1:-1]
        assert np.allclose(fuels_dens_start, self.line_fire_sim.quic_fire.fuel_density)

        # Load the next time step
        fuels_dens_end = _process_compressed_bin(
            LINE_FIRE_DIR / "Output" / "fuels-dens-00060.bin",
            (
                self.line_fire_sim.quic_fire.nz,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
            self.line_fire_indexes,
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "fuels-dens-00060.npz")
        reference_fuel_dens = npz["var"]
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 0, 2)
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 1, 2)

        assert np.allclose(fuels_dens_end[0, 0, ...], reference_fuel_dens)

    def test_eglin_fuels_dens(self):
        # Load the first time step model output of fuel density
        fuels_dens_start = _process_compressed_bin(
            EG_CANOPY_DIR / "Output" / "fuels-dens-00000.bin",
            (
                self.eglin_canopy_sim.quic_fire.nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
            self.eglin_canopy_indexes,
        )
        assert fuels_dens_start.shape == (
            1,
            self.eglin_canopy_sim.quic_fire.nz,
            self.eglin_canopy_sim.qu_simparams.ny,
            self.eglin_canopy_sim.qu_simparams.nx,
        )

        # Load in the input .dat file for comparison
        fuels_dens_dat = read_dat_file(
            EG_CANOPY_DIR / "treesrhof.dat",
            self.eglin_canopy_sim.quic_fire.nz,
            self.eglin_canopy_sim.qu_simparams.ny,
            self.eglin_canopy_sim.qu_simparams.nx,
        )

        # I have no idea why the edges are wonky, but the rest of the array
        # is fine. Maybe there's some model interpolation going on?
        assert np.allclose(
            fuels_dens_start[0, :, 1:-1, 2:-1], fuels_dens_dat[:, 1:-1, 1:-2], atol=1e-3
        )

        # Load the first time step of the reference data
        npz = np.load(EG_CANOPY_DIR / "Reference" / "fuels-dens-00000.npz")
        reference_fuel_dens = npz["var"]
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 0, 2)
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 1, 2)

        assert np.allclose(fuels_dens_start[0, ...], reference_fuel_dens)

        # Load the next time step
        fuels_dens_end = _process_compressed_bin(
            EG_CANOPY_DIR / "Output" / "fuels-dens-00600.bin",
            (
                self.eglin_canopy_sim.quic_fire.nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
            self.eglin_canopy_indexes,
        )

        # Load the next time step of the reference data
        npz = np.load(EG_CANOPY_DIR / "Reference" / "fuels-dens-00600.npz")
        reference_fuel_dens = npz["var"]
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 0, 2)
        reference_fuel_dens = np.swapaxes(reference_fuel_dens, 1, 2)

        assert np.allclose(fuels_dens_end[0, ...], reference_fuel_dens)

    def test_eglin_co_emissions(self):
        # Load the model output
        co_emissions = _process_compressed_bin(
            EG_CANOPY_DIR / "Output" / "co_emissions-00600.bin",
            (
                self.eglin_canopy_sim.quic_fire.nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
            self.eglin_canopy_indexes,
        )

        # Load the reference data
        npz = np.load(EG_CANOPY_DIR / "Reference" / "co_emissions-00600.npz")
        reference_co_emissions = npz["var"]
        reference_co_emissions = np.swapaxes(reference_co_emissions, 0, 2)
        reference_co_emissions = np.swapaxes(reference_co_emissions, 1, 2)

        assert np.allclose(co_emissions[0, ...], reference_co_emissions)

    def test_eglin_thermal_radiation(self):
        # Load the model output
        thermal_radiation = _process_compressed_bin(
            EG_CANOPY_DIR / "Output" / "thermalradiation-00600.bin",
            (
                self.eglin_canopy_sim.quic_fire.nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
            self.eglin_canopy_indexes,
        )

        # Load the reference data
        npz = np.load(EG_CANOPY_DIR / "Reference" / "thermalradiation-00600.npz")
        reference_thermal_radiation = npz["var"]
        reference_thermal_radiation = np.swapaxes(reference_thermal_radiation, 0, 2)
        reference_thermal_radiation = np.swapaxes(reference_thermal_radiation, 1, 2)

        assert np.allclose(thermal_radiation[0, ...], reference_thermal_radiation)


class TestProcessGriddedBin:
    line_fire_sim = SimulationInputs.from_directory(LINE_FIRE_DIR)
    line_fire_indexes = _process_fire_indexes(
        LINE_FIRE_DIR / "Output" / "fire_indexes.bin"
    )

    eglin_canopy_sim = SimulationInputs.from_directory(EG_CANOPY_DIR)
    eglin_canopy_indexes = _process_fire_indexes(
        EG_CANOPY_DIR / "Output" / "fire_indexes.bin"
    )

    def test_line_fire_eng2atmos(self):
        _, _, line_fire_eng2atmos_nz, _ = _process_grid_info(
            LINE_FIRE_DIR / "Output" / "grid.bin",
            self.line_fire_sim.qu_simparams.ny,
            self.line_fire_sim.qu_simparams.nx,
        )

        # Load first time step
        eng2atmos = _process_gridded_bin(
            LINE_FIRE_DIR / "Output" / "fire-energy_to_atmos-00000.bin",
            (
                line_fire_eng2atmos_nz,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "fire-energy_to_atmos-00000.npz")
        reference_eng2atmos = npz["var"]
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 0, 2)
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 1, 2)

        assert np.allclose(eng2atmos[0], reference_eng2atmos)

        # Load next time step
        eng2atmos = _process_gridded_bin(
            LINE_FIRE_DIR / "Output" / "fire-energy_to_atmos-00060.bin",
            (
                line_fire_eng2atmos_nz,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
            self.line_fire_indexes,
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "fire-energy_to_atmos-00060.npz")
        reference_eng2atmos = npz["var"]
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 0, 2)
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 1, 2)

        assert np.allclose(eng2atmos[0], reference_eng2atmos)

    def test_line_fire_surf_energy(self):
        # Load first time step
        surf_energy = _process_gridded_bin(
            LINE_FIRE_DIR / "Output" / "surfEnergy00009.bin",
            (
                1,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
        )

        assert surf_energy.shape == (1, 1, 200, 200)
        assert surf_energy.max() > 0

    def test_line_fire_mburnt_integ(self):
        # Load first time step
        mburnt_integ = _process_gridded_bin(
            LINE_FIRE_DIR / "Output" / "mburnt_integ-00000.bin",
            (
                1,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "mburnt_integ-00000.npz")
        reference_mburnt_integ = npz["var"]
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 0, 2)
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 1, 2)

        assert np.allclose(mburnt_integ[0], reference_mburnt_integ)

        # Load next time step
        mburnt_integ = _process_gridded_bin(
            LINE_FIRE_DIR / "Output" / "mburnt_integ-00060.bin",
            (
                1,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
            self.line_fire_indexes,
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "mburnt_integ-00060.npz")
        reference_mburnt_integ = npz["var"]
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 0, 2)
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 1, 2)

        assert np.allclose(mburnt_integ[0], reference_mburnt_integ)

    def test_line_fire_groundfuelheight(self):
        # Load model output
        groundfuelheight = _process_gridded_bin(
            LINE_FIRE_DIR / "Output" / "groundfuelheight.bin",
            (
                1,
                self.line_fire_sim.qu_simparams.ny,
                self.line_fire_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(LINE_FIRE_DIR / "Reference" / "groundfuelheight.npz")
        reference_groundfuelheight = npz["var"]

        assert np.allclose(groundfuelheight[0], reference_groundfuelheight)

    def test_eglin_eng2atmos(self):
        _, _, eglin_eng2atmos_nz, _ = _process_grid_info(
            EG_CANOPY_DIR / "Output" / "grid.bin",
            self.eglin_canopy_sim.qu_simparams.ny,
            self.eglin_canopy_sim.qu_simparams.nx,
        )

        # Load first time step
        eng2atmos = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "fire-energy_to_atmos-00000.bin",
            (
                eglin_eng2atmos_nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "fire-energy_to_atmos-00000.npz")
        reference_eng2atmos = npz["var"]
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 0, 2)
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 1, 2)

        assert eng2atmos[0].shape == reference_eng2atmos.shape
        assert np.allclose(eng2atmos[0], reference_eng2atmos)

        # Load next time step
        eng2atmos = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "fire-energy_to_atmos-00600.bin",
            (
                eglin_eng2atmos_nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
            self.eglin_canopy_indexes,
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "fire-energy_to_atmos-00600.npz")
        reference_eng2atmos = npz["var"]
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 0, 2)
        reference_eng2atmos = np.swapaxes(reference_eng2atmos, 1, 2)

        assert eng2atmos[0].shape == reference_eng2atmos.shape
        assert np.allclose(eng2atmos[0], reference_eng2atmos)

    def test_eglin_qu_windu(self):
        # Load first time step
        qu_windu = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "qu_windu00600.bin",
            (
                self.eglin_canopy_sim.qu_simparams.nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "qu_windu00600.npz")
        reference_qu_windu = npz["var"]
        reference_qu_windu = np.swapaxes(reference_qu_windu, 0, 2)
        reference_qu_windu = np.swapaxes(reference_qu_windu, 1, 2)
        reference_qu_windu = reference_qu_windu[:-1, ...]

        assert np.allclose(qu_windu[0], reference_qu_windu)

    def test_eglin_windu(self):
        # Load first time step
        windu = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "windu00600.bin",
            (
                self.eglin_canopy_sim.quic_fire.nz,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "windu00600.npz")
        reference_windu = npz["var"]
        reference_windu = np.swapaxes(reference_windu, 0, 2)
        reference_windu = np.swapaxes(reference_windu, 1, 2)
        reference_windu = reference_windu[:-1, ...]

        assert np.allclose(windu[0], reference_windu)

    def test_eglin_mburnt_integ(self):
        # Load first time step
        mburnt_integ = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "mburnt_integ-00000.bin",
            (
                1,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "mburnt_integ-00000.npz")
        reference_mburnt_integ = npz["var"]
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 0, 2)
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 1, 2)

        assert np.allclose(mburnt_integ[0], reference_mburnt_integ)

        # Load next time step
        mburnt_integ = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "mburnt_integ-00600.bin",
            (
                1,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "mburnt_integ-00600.npz")
        reference_mburnt_integ = npz["var"]
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 0, 2)
        reference_mburnt_integ = np.swapaxes(reference_mburnt_integ, 1, 2)

        assert np.allclose(mburnt_integ[0], reference_mburnt_integ)

    def test_eglin_groundfuelheight(self):
        # Load model output
        groundfuelheight = _process_gridded_bin(
            EG_CANOPY_DIR / "Output" / "groundfuelheight.bin",
            (
                1,
                self.eglin_canopy_sim.qu_simparams.ny,
                self.eglin_canopy_sim.qu_simparams.nx,
            ),
        )

        # Load reference data for comparison
        npz = np.load(EG_CANOPY_DIR / "Reference" / "groundfuelheight.npz")
        reference_groundfuelheight = npz["var"]

        assert np.allclose(groundfuelheight[0], reference_groundfuelheight)


class TestSimulationOutputs:
    def test_line_fire(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        outputs = SimulationOutputs(
            LINE_FIRE_DIR / "Output",
            line_fire.quic_fire.nz,
            line_fire.qu_simparams.ny,
            line_fire.qu_simparams.nx,
            line_fire.qu_simparams.dy,
            line_fire.qu_simparams.dx,
        )

        # Check that the grid info is correct
        assert outputs.fire_nz == line_fire.quic_fire.nz
        assert outputs.quic_nz == line_fire.qu_simparams.nz
        assert outputs.fire_dz == line_fire.quic_fire.dz
        assert np.allclose(outputs.quic_dz, line_fire.qu_simparams._dz_array[:-1])

    def test_line_fire_georeferenced(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        utm_zones = range(1, 16)
        epsg_codes = [32600 + utm_zone for utm_zone in utm_zones]
        for zone, epsg in zip(utm_zones, epsg_codes):
            outputs = SimulationOutputs(
                LINE_FIRE_DIR / "Output",
                line_fire.quic_fire.nz,
                line_fire.qu_simparams.ny,
                line_fire.qu_simparams.nx,
                line_fire.qu_simparams.dy,
                line_fire.qu_simparams.dx,
                utm_x=np.random.randint(1, 1e8),
                utm_y=np.random.randint(1, 1e8),
                utm_zone=zone,
            )

            assert outputs.crs == f"EPSG:{epsg}"

    def test_line_fire_list_outputs(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        outputs = SimulationOutputs(
            LINE_FIRE_DIR / "Output",
            line_fire.quic_fire.nz,
            line_fire.qu_simparams.ny,
            line_fire.qu_simparams.nx,
            line_fire.qu_simparams.dy,
            line_fire.qu_simparams.dx,
        )
        outputs_list = outputs.list_outputs()

        assert "fire-energy_to_atmos" in outputs_list
        assert "fuels-dens" in outputs_list
        assert "groundfuelheight" in outputs_list
        assert "mburnt_integ" in outputs_list
        assert "surfEnergy" in outputs_list

    def test_line_fire_get_output(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        outputs = SimulationOutputs(
            LINE_FIRE_DIR / "Output",
            line_fire.quic_fire.nz,
            line_fire.qu_simparams.ny,
            line_fire.qu_simparams.nx,
            line_fire.qu_simparams.dy,
            line_fire.qu_simparams.dx,
        )

        fire_energy = outputs.get_output("fire-energy_to_atmos")
        assert fire_energy.times == [0, 60]

        fuels_dens = outputs.get_output("fuels-dens")
        assert fuels_dens.times == [0, 60]

        groundfuelheight = outputs.get_output("groundfuelheight")
        assert groundfuelheight.times == [0]

        mburnt_integ = outputs.get_output("mburnt_integ")
        assert mburnt_integ.times == [0, 60]

        surf_energy = outputs.get_output("surfEnergy")
        assert surf_energy.times == [0, 9]

    def test_line_fire_from_simulation(self):
        # Create the SimulationOutputs object with the normal constructor
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        outputs = SimulationOutputs(
            LINE_FIRE_DIR / "Output",
            line_fire.quic_fire.nz,
            line_fire.qu_simparams.ny,
            line_fire.qu_simparams.nx,
            line_fire.qu_simparams.dy,
            line_fire.qu_simparams.dx,
        )

        # Create the SimulationOutputs object from the SimulationInputs object
        outputs_from_sim_inputs = SimulationOutputs.from_simulation_inputs(
            LINE_FIRE_DIR / "Output", line_fire
        )

        # The two objects should be the same
        assert outputs == outputs_from_sim_inputs

    def test_eglin_canopy(self):
        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        outputs = SimulationOutputs(
            EG_CANOPY_DIR / "Output",
            eglin_canopy.quic_fire.nz,
            eglin_canopy.qu_simparams.ny,
            eglin_canopy.qu_simparams.nx,
            eglin_canopy.qu_simparams.dy,
            eglin_canopy.qu_simparams.dx,
        )

        # Check that the grid info is correct
        assert outputs.fire_nz == eglin_canopy.quic_fire.nz
        assert outputs.quic_nz == eglin_canopy.qu_simparams.nz
        assert outputs.fire_dz == eglin_canopy.quic_fire.dz
        assert np.allclose(outputs.quic_dz, eglin_canopy.qu_simparams._dz_array[:-1])

    def test_eglin_canopy_list_outputs(self):
        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        outputs = SimulationOutputs(
            EG_CANOPY_DIR / "Output",
            eglin_canopy.quic_fire.nz,
            eglin_canopy.qu_simparams.ny,
            eglin_canopy.qu_simparams.nx,
            eglin_canopy.qu_simparams.dy,
            eglin_canopy.qu_simparams.dx,
        )

        assert "co_emissions" in outputs.list_outputs()
        assert "fire-energy_to_atmos" in outputs.list_outputs()
        assert "fuels-dens" in outputs.list_outputs()
        assert "groundfuelheight" in outputs.list_outputs()
        assert "mburnt_integ" in outputs.list_outputs()
        assert "qu_windu" in outputs.list_outputs()
        assert "thermalradiation" in outputs.list_outputs()
        assert "windu" in outputs.list_outputs()

    def test_eglin_canopy_get_output(self):
        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        outputs = SimulationOutputs(
            EG_CANOPY_DIR / "Output",
            eglin_canopy.quic_fire.nz,
            eglin_canopy.qu_simparams.ny,
            eglin_canopy.qu_simparams.nx,
            eglin_canopy.qu_simparams.dy,
            eglin_canopy.qu_simparams.dx,
        )

        co_emissions = outputs.get_output("co_emissions")
        assert co_emissions.times == [600]
        assert co_emissions.shape == (outputs.fire_nz, outputs.ny, outputs.nx)

        fire_energy = outputs.get_output("fire-energy_to_atmos")
        assert fire_energy.times == [0, 600]
        assert fire_energy.shape == (outputs.en2atmos_nz, outputs.ny, outputs.nx)

        groundfuelheight = outputs.get_output("groundfuelheight")
        assert groundfuelheight.times == [0]
        assert groundfuelheight.shape == (1, outputs.ny, outputs.nx)

        mburnt_integ = outputs.get_output("mburnt_integ")
        assert mburnt_integ.times == [0, 600]
        assert mburnt_integ.shape == (1, outputs.ny, outputs.nx)

        qu_windu = outputs.get_output("qu_windu")
        assert qu_windu.times == [600]
        assert qu_windu.shape == (outputs.quic_nz, outputs.ny, outputs.nx)

        thermal_rad = outputs.get_output("thermalradiation")
        assert thermal_rad.times == [600]
        assert thermal_rad.shape == (outputs.fire_nz, outputs.ny, outputs.nx)

        wind = outputs.get_output("windu")
        assert wind.times == [600]
        assert wind.shape == (outputs.fire_nz, outputs.ny, outputs.nx)

    def test_object_equality(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        line_fire_outputs = SimulationOutputs(
            LINE_FIRE_DIR / "Output",
            line_fire.quic_fire.nz,
            line_fire.qu_simparams.ny,
            line_fire.qu_simparams.nx,
            line_fire.qu_simparams.dy,
            line_fire.qu_simparams.dx,
        )

        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        eglin_canopy_outputs = SimulationOutputs(
            EG_CANOPY_DIR / "Output",
            eglin_canopy.quic_fire.nz,
            eglin_canopy.qu_simparams.ny,
            eglin_canopy.qu_simparams.nx,
            eglin_canopy.qu_simparams.dy,
            eglin_canopy.qu_simparams.dx,
        )

        assert line_fire_outputs != eglin_canopy_outputs


class TestOutputFileToNumpy:
    line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
    line_fire_outputs = SimulationOutputs.from_simulation_inputs(
        LINE_FIRE_DIR / "Output", line_fire
    )

    eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
    eglin_canopy_outputs = SimulationOutputs.from_simulation_inputs(
        EG_CANOPY_DIR / "Output", eglin_canopy
    )

    @staticmethod
    def _test_to_numpy(output: OutputFile, times, nz, ny, nx):
        # Get all time steps
        data_all = output.to_numpy()
        assert data_all.shape == (len(times), nz, ny, nx)

        # Get all time steps with range
        data_all = output.to_numpy(range(len(times)))
        assert data_all.shape == (len(times), nz, ny, nx)

        # Get every other time step with range
        data_all = output.to_numpy(range(0, len(times), 2))
        assert data_all.shape == (max(len(times) // 2, 1), nz, ny, nx)

        # Get the first time step
        data_first = output.to_numpy(timestep=0)
        assert data_first.shape == (1, nz, ny, nx)

        # Get the last time step
        data_last = output.to_numpy(timestep=-1)
        assert data_last.shape == (1, nz, ny, nx)

    def test_lf_eng2atmos(self):
        eng2atmos = self.line_fire_outputs.get_output("fire-energy_to_atmos")
        self._test_to_numpy(eng2atmos, eng2atmos.times, *eng2atmos.shape)

    def test_lf_fuels_dens(self):
        fuels_dens = self.line_fire_outputs.get_output("fuels-dens")
        self._test_to_numpy(fuels_dens, fuels_dens.times, *fuels_dens.shape)

    def test_lf_groundfuelheight(self):
        groundfuelheight = self.line_fire_outputs.get_output("groundfuelheight")
        self._test_to_numpy(
            groundfuelheight,
            groundfuelheight.times,
            *groundfuelheight.shape,
        )

    def test_lf_mburnt_integ(self):
        mburnt_integ = self.line_fire_outputs.get_output("mburnt_integ")
        self._test_to_numpy(
            mburnt_integ,
            mburnt_integ.times,
            *mburnt_integ.shape,
        )

    def test_lf_surf_energy(self):
        surf_energy = self.line_fire_outputs.get_output("surfEnergy")
        self._test_to_numpy(
            surf_energy,
            surf_energy.times,
            *surf_energy.shape,
        )

    def test_ec_co_emissions(self):
        co_emissions = self.eglin_canopy_outputs.get_output("co_emissions")
        self._test_to_numpy(co_emissions, co_emissions.times, *co_emissions.shape)

    def test_ec_eng2atmos(self):
        eng2atmos = self.eglin_canopy_outputs.get_output("fire-energy_to_atmos")
        self._test_to_numpy(eng2atmos, eng2atmos.times, *eng2atmos.shape)

    def test_ec_fuel_dens(self):
        fuels_dens = self.eglin_canopy_outputs.get_output("fuels-dens")
        self._test_to_numpy(fuels_dens, fuels_dens.times, *fuels_dens.shape)

    def test_ec_groundfuelheight(self):
        groundfuelheight = self.eglin_canopy_outputs.get_output("groundfuelheight")
        self._test_to_numpy(
            groundfuelheight,
            groundfuelheight.times,
            *groundfuelheight.shape,
        )

    def test_ec_mburnt_integ(self):
        mburnt_integ = self.eglin_canopy_outputs.get_output("mburnt_integ")
        self._test_to_numpy(
            mburnt_integ,
            mburnt_integ.times,
            *mburnt_integ.shape,
        )

    def test_ec_qu_windu_to_numpy(self):
        qu_windu = self.eglin_canopy_outputs.get_output("qu_windu")
        self._test_to_numpy(
            qu_windu,
            qu_windu.times,
            *qu_windu.shape,
        )

    def test_ec_thermal_radiation_to_numpy(self):
        thermal_radiation = self.eglin_canopy_outputs.get_output("thermalradiation")
        self._test_to_numpy(
            thermal_radiation,
            thermal_radiation.times,
            *thermal_radiation.shape,
        )

    def test_ec_windu_to_numpy(self):
        windu = self.eglin_canopy_outputs.get_output("windu")
        self._test_to_numpy(
            windu,
            windu.times,
            *windu.shape,
        )


class TestOutputFileToNetCDF:
    line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
    line_fire_outputs = SimulationOutputs.from_simulation_inputs(
        LINE_FIRE_DIR / "Output", line_fire
    )

    eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
    eglin_canopy_outputs = SimulationOutputs.from_simulation_inputs(
        EG_CANOPY_DIR / "Output", eglin_canopy
    )

    @staticmethod
    def _test_to_netcdf(output: OutputFile, directory: Path, times, nz, ny, nx):
        # Get all time steps
        data_all = output.to_numpy()
        output.to_netcdf(directory)
        test = nc.Dataset(directory / f"{output.name}.nc", "r")
        data_all = test.variables[output.name]
        assert data_all.shape == (len(times), nz, ny, nx)
        test.close()

        # Get all time steps with range
        data_all = output.to_numpy(range(len(times)))
        output.to_netcdf(directory, range(len(times)))
        test = nc.Dataset(directory / f"{output.name}.nc", "r")
        data_all = test.variables[output.name]
        assert data_all.shape == (len(times), nz, ny, nx)
        test.close()

        # Get every other time step with range
        data_all = output.to_numpy(range(0, len(times), 2))
        output.to_netcdf(directory, range(0, len(times), 2))
        test = nc.Dataset(directory / f"{output.name}.nc", "r")
        data_all = test.variables[output.name]
        assert data_all.shape == (max(len(times) // 2, 1), nz, ny, nx)
        test.close()

        # Get the first time step
        data_first = output.to_numpy(timestep=0)
        output.to_netcdf(directory, timestep=0)
        test = nc.Dataset(directory / f"{output.name}.nc", "r")
        data_first = test.variables[output.name]
        assert data_first.shape == (1, nz, ny, nx)
        test.close()

        # Get the last time step
        data_last = output.to_numpy(timestep=-1)
        output.to_netcdf(directory, timestep=-1)
        test = nc.Dataset(directory / f"{output.name}.nc", "r")
        data_last = test.variables[output.name]
        assert data_last.shape == (1, nz, ny, nx)
        test.close()

    def test_lf_eng2atmos(self):
        eng2atmos = self.line_fire_outputs.get_output("fire-energy_to_atmos")
        self._test_to_netcdf(eng2atmos, TMP_DIR, eng2atmos.times, *eng2atmos.shape)

    def test_lf_fuels_dens(self):
        fuels_dens = self.line_fire_outputs.get_output("fuels-dens")
        self._test_to_netcdf(fuels_dens, TMP_DIR, fuels_dens.times, *fuels_dens.shape)

    def test_lf_groundfuelheight(self):
        groundfuelheight = self.line_fire_outputs.get_output("groundfuelheight")
        self._test_to_netcdf(
            groundfuelheight,
            TMP_DIR,
            groundfuelheight.times,
            *groundfuelheight.shape,
        )

    def test_lf_mburnt_integ(self):
        mburnt_integ = self.line_fire_outputs.get_output("mburnt_integ")
        self._test_to_netcdf(
            mburnt_integ,
            TMP_DIR,
            mburnt_integ.times,
            *mburnt_integ.shape,
        )

    def test_lf_surf_energy(self):
        surf_energy = self.line_fire_outputs.get_output("surfEnergy")
        self._test_to_netcdf(
            surf_energy,
            TMP_DIR,
            surf_energy.times,
            *surf_energy.shape,
        )

    def test_ec_co_emissions(self):
        co_emissions = self.eglin_canopy_outputs.get_output("co_emissions")
        self._test_to_netcdf(
            co_emissions, TMP_DIR, co_emissions.times, *co_emissions.shape
        )

    def test_ec_eng2atmos(self):
        eng2atmos = self.eglin_canopy_outputs.get_output("fire-energy_to_atmos")
        self._test_to_netcdf(eng2atmos, TMP_DIR, eng2atmos.times, *eng2atmos.shape)

    def test_ec_fuel_dens(self):
        fuels_dens = self.eglin_canopy_outputs.get_output("fuels-dens")
        self._test_to_netcdf(fuels_dens, TMP_DIR, fuels_dens.times, *fuels_dens.shape)

    def test_ec_groundfuelheight(self):
        groundfuelheight = self.eglin_canopy_outputs.get_output("groundfuelheight")
        self._test_to_netcdf(
            groundfuelheight,
            TMP_DIR,
            groundfuelheight.times,
            *groundfuelheight.shape,
        )

    def test_ec_mburnt_integ(self):
        mburnt_integ = self.eglin_canopy_outputs.get_output("mburnt_integ")
        self._test_to_netcdf(
            mburnt_integ,
            TMP_DIR,
            mburnt_integ.times,
            *mburnt_integ.shape,
        )

    def test_ec_qu_windu_to_numpy(self):
        qu_windu = self.eglin_canopy_outputs.get_output("qu_windu")
        self._test_to_netcdf(
            qu_windu,
            TMP_DIR,
            qu_windu.times,
            *qu_windu.shape,
        )

    def test_ec_thermal_radiation_to_numpy(self):
        thermal_radiation = self.eglin_canopy_outputs.get_output("thermalradiation")
        self._test_to_netcdf(
            thermal_radiation,
            TMP_DIR,
            thermal_radiation.times,
            *thermal_radiation.shape,
        )

    def test_ec_windu_to_numpy(self):
        windu = self.eglin_canopy_outputs.get_output("windu")
        self._test_to_netcdf(
            windu,
            TMP_DIR,
            windu.times,
            *windu.shape,
        )


class TestOutputFileToZarr:
    def test_lf_fuel_dens(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        line_fire_outputs = SimulationOutputs.from_simulation_inputs(
            LINE_FIRE_DIR / "Output", line_fire
        )

        fuels_dens = line_fire_outputs.get_output("fuels-dens")
        fuels_dens.to_zarr(TMP_DIR)

        zroot = zarr.open(str(TMP_DIR / "fuels-dens.zarr"), mode="r")
        assert np.allclose(zroot["time"][:], fuels_dens.times)
        assert np.allclose(zroot["x"][:], fuels_dens.x_coords)
        assert np.allclose(zroot["y"][:], fuels_dens.y_coords)
        assert np.allclose(zroot["z"][:], fuels_dens.z_coords)

        z_array = zroot["fuels-dens"]
        assert np.allclose(z_array[:], fuels_dens.to_numpy())
        assert z_array.chunks == (1, *fuels_dens.shape)

    def test_lf_fuel_dens_custom_time_chunks(self):
        line_fire = SimulationInputs.from_directory(LINE_FIRE_DIR)
        line_fire_outputs = SimulationOutputs.from_simulation_inputs(
            LINE_FIRE_DIR / "Output", line_fire
        )

        fuels_dens = line_fire_outputs.get_output("fuels-dens")
        fuels_dens.to_zarr(TMP_DIR, chunk_size={"time": 100})

        zroot = zarr.open(str(TMP_DIR / "fuels-dens.zarr"), mode="r")
        z_array = zroot["fuels-dens"]
        print(z_array.chunks)

        assert z_array.chunks == (100, *fuels_dens.shape)

    def test_ec_qu_windu(self):
        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        eglin_canopy_outputs = SimulationOutputs.from_simulation_inputs(
            EG_CANOPY_DIR / "Output", eglin_canopy
        )

        qu_windu = eglin_canopy_outputs.get_output("qu_windu")
        qu_windu.to_zarr(TMP_DIR)

        zroot = zarr.open(str(TMP_DIR / "qu_windu.zarr"), mode="r")
        assert np.allclose(zroot["time"][:], qu_windu.times)
        assert np.allclose(zroot["x"][:], qu_windu.x_coords)
        assert np.allclose(zroot["y"][:], qu_windu.y_coords)
        assert np.allclose(zroot["z"][:], qu_windu.z_coords)

        z_array = zroot["qu_windu"]
        assert np.allclose(z_array[:], qu_windu.to_numpy())
        assert z_array.chunks == (1, *qu_windu.shape)

    def test_ec_qu_windu_custom_chunks(self):
        eglin_canopy = SimulationInputs.from_directory(EG_CANOPY_DIR)
        eglin_canopy_outputs = SimulationOutputs.from_simulation_inputs(
            EG_CANOPY_DIR / "Output", eglin_canopy
        )

        qu_windu = eglin_canopy_outputs.get_output("qu_windu")
        qu_windu.to_zarr(
            TMP_DIR,
            chunk_size={
                "time": 10,
                "z": qu_windu.shape[0] // 2,
                "y": qu_windu.shape[1] // 2,
                "x": qu_windu.shape[2] // 2,
            },
        )

        zroot = zarr.open(str(TMP_DIR / "qu_windu.zarr"), mode="r")
        z_array = zroot["qu_windu"]

        assert z_array.chunks == (
            10,
            qu_windu.shape[0] // 2,
            qu_windu.shape[1] // 2,
            qu_windu.shape[2] // 2,
        )
        assert z_array.cdata_shape == (1, 2, 2, 2)



"""
QUIC-Fire Tools Topography Model
"""

from __future__ import annotations

# Core Imports
from enum import Enum

# External Imports
from typing import Literal
from pydantic import BaseModel, Field, PositiveFloat, SerializeAsAny


class TopoFlags(int, Enum):
    """
    Enum class for all valid topography source options in QUIC-Fire.
    """

    flat = 0
    gaussian_hill = 1
    hill_pass = 2
    slope_mesa = 3
    canyon = 4
    custom = 5
    half_circle = 6
    sinusoid = 7
    cos_hill = 8
    QP_elevation_bin = 9
    terrainOutput_txt = 10
    terrain_dat = 11


class Topography(BaseModel):
    """
    Base class for all topography types in QUIC-Fire. This class is used to
    provide a common string representation for all topography types.
    """

    topo_flag: SerializeAsAny[TopoFlags]

    def __str__(self):
        return (
            f"{self.topo_flag.value}\t\t! N/A, "
            f"topo flag: 0 = flat, 1 = Gaussian hill, "
            f"2 = hill pass, 3 = slope mesa, 4 = canyon, "
            f"5 = custom, 6 = half circle, 7 = sinusoid, "
            f"8 = cos hill, 9 = QP_elevation.inp, "
            f"10 = terrainOutput.txt (ARA), "
            f"11 = terrain.dat (firetec)"
        )


class GaussianHillTopo(Topography):
    """
    Creates a Gaussian hill topography in QUIC-Fire.

    Attributes
    ----------
    x_hilltop : PositiveFloat
        Gaussian hill top location x-direction [m]
    y_hilltop : PositiveFloat
        Gaussian hill top location y-direction [m]
    elevation_max : PositiveFloat
        Maximum elevation of the Gaussian hill [m]
    elevation_std : PositiveFloat
        Standard deviation of the Gaussian hill [m]
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(1)
    x_hilltop: PositiveFloat
    y_hilltop: PositiveFloat
    elevation_max: PositiveFloat
    elevation_std: PositiveFloat

    def __str__(self):
        flag_line = super().__str__()
        params = (
            f"\n{self.x_hilltop}\t! m, x-center\n"
            f"{self.y_hilltop}\t! m, y-center\n"
            f"{self.elevation_max}\t! m, max height\n"
            f"{self.elevation_std}\t! m, std"
        )
        return flag_line + params


class HillPassTopo(Topography):
    """
    Creates a hill pass topography in QUIC-Fire.

    Attributes
    ----------
    max_height : PositiveFloat
        Maximum elevation of the hill pass [m]
    location_param : PositiveFloat
        Location parameter of the hill pass [m]
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(2)
    max_height: PositiveFloat
    location_param: PositiveFloat

    def __str__(self):
        flag_line = super().__str__()
        params = (
            f"\n{self.max_height}\t! m, max height\n"
            f"{self.location_param}\t! m, location parameter"
        )
        return flag_line + params


class SlopeMesaTopo(Topography):
    """
    Creates a slope mesa topography in QUIC-Fire.

    Attributes
    ----------
    slope_axis : Literal[0, 1]
        Slope axis (0 = x, 1 = y)
    slope_value : float
        Slope in dh/dx or dh/dy
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(3)
    slope_axis: Literal[0, 1]
    slope_value: PositiveFloat
    flat_fraction: float = Field(ge=0, le=1)

    def __str__(self):
        flag_line = super().__str__()
        params = (
            f"\n{self.slope_axis}\t! N/A, slope axis (0 = x, 1 = y)\n"
            f"{self.slope_value}\t! N/A, slope val (dh/dx)\n"
            f"{self.flat_fraction}\t! N/A, fraction of domain that is flat"
        )
        return flag_line + params


class CanyonTopo(Topography):
    """
    Creates a canyon topography in QUIC-Fire.

    Attributes
    ----------
    x_location: PositiveFloat
        Canyon location in x-dir [m].
    y_location: PositiveFloat
        Canyon location in y-dir [m].
    slope_value: PositiveFloat
        Slope in dh/dx or dy/dy [-].
    canyon_std: PositiveFloat
        Canyon function standard deviation [m].
    vertical_offset: PositiveFloat
        Canyon vertical offset [m].
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(4)
    x_location: PositiveFloat
    y_location: PositiveFloat
    slope_value: PositiveFloat
    canyon_std: PositiveFloat
    vertical_offset: PositiveFloat

    def __str__(self):
        flag_line = super().__str__()
        params = (
            f"\n{self.x_location}\t! m, x-start of canyon on slope\n"
            f"{self.y_location}\t! m, y-center of canyon"
            f"{self.slope_value}\t! N/A, slope\n"
            f"{self.canyon_std}\t! m, std\n"
            f"{self.vertical_offset}\t! m, height offset of the bottom of the canyon"
        )
        return flag_line + params


class HalfCircleTopo(Topography):
    """
    Creates a half-circle topography in QUIC-Fire.

    Attributes
    ----------
    x_location: PositiveFloat
        The x-coordinate of the center of the half-circle topography [m].
    y_location: PositiveFloat
        The y-coordinate of the center of the half-circle topography [m].
    radius: PositiveFloat
        The radius of the half-circle topography [m].
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(6)
    x_location: PositiveFloat
    y_location: PositiveFloat
    radius: PositiveFloat

    def __str__(self):
        flag_line = super().__str__()
        params = (
            f"\n{self.x_location}\t! m, x-location\n"
            f"{self.y_location}\t! m, y-location\n"
            f"{self.radius}\t! m, radius of half circle"
        )
        return flag_line + params


class SinusoidTopo(Topography):
    """
    Creates a sinusoidal topography in QUIC-Fire.

    Attributes
    ----------
    period: PositiveFloat
        The period of the sinusoidal wave [m].
    amplitude: PositiveFloat
        The amplitude of the sinusoidal wave [m].
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(7)
    period: PositiveFloat
    amplitude: PositiveFloat

    def __str__(self):
        flag_line = super().__str__()
        params = f"\n{self.period}\t! m, period\n" f"{self.amplitude}\t! m, amplitude"
        return flag_line + params


class CosHillTopo(Topography):
    """
    Creates a cosine-shaped hill topography in QUIC-Fire.

    Attributes
    ----------
    aspect: PositiveFloat
        The aspect (or orientation) of the hill [-].
    height: PositiveFloat
        The height of the hill [m].
    """

    topo_flag: SerializeAsAny[TopoFlags] = TopoFlags(8)
    aspect: PositiveFloat
    height: PositiveFloat

    def __str__(self):
        flag_line = super().__str__()
        params = f"\n{self.aspect}\t! [0], aspect\n" f"{self.height}\t! m, height"
        return flag_line + params


def serialize_topography(topo_data: dict):
    topo_flag = topo_data.get("topo_flag")
    if topo_flag == TopoFlags(1):
        return GaussianHillTopo(**topo_data)
    elif topo_flag == TopoFlags(2):
        return HillPassTopo(**topo_data)
    elif topo_flag == TopoFlags(3):
        return SlopeMesaTopo(**topo_data)
    elif topo_flag == TopoFlags(4):
        return CanyonTopo(**topo_data)
    elif topo_flag == TopoFlags(6):
        return HalfCircleTopo(**topo_data)
    elif topo_flag == TopoFlags(7):
        return SinusoidTopo(**topo_data)
    elif topo_flag == TopoFlags(8):
        return CosHillTopo(**topo_data)
    else:
        return Topography(**topo_data)



"""
Module for converting QUIC-Fire output files to duck array data formats.
"""

from __future__ import annotations

# Core imports
import re
import json
from pathlib import Path
import importlib.resources

# Internal imports
from quicfire_tools.inputs import SimulationInputs

# External imports
import zarr
import numpy as np
import pyproj as proj
import dask.array as da
from numpy import ndarray
from netCDF4 import Dataset


OUTPUTS_DIR_PATH = (
    importlib.resources.files("quicfire_tools").joinpath("data").joinpath("outputs")
)
with open(OUTPUTS_DIR_PATH.joinpath("outputs.json")) as f:
    OUTPUTS_MAP = json.load(f)


class OutputFile:
    """
    A class representing a single output file. This class provides a common
    interface for processing and retrieving data from output files.

    Attributes
    ----------
    name: str
        The name of the output file this object represents.
    file_format: str
        The format of the output file. Valid options are "gridded" and
        "compressed".
    dimensions: list[str]
        The dimensions of the output file as ["z", "y", "x"] or ["y", "x"].
    shape: tuple
        The shape of the output file array data as (nz, ny, nx).
    grid: str
        The grid type of the output file. Valid options are "fire" and "quic".
    x_coords: list[float]
        The x coordinates of the output file. Coordinates are in the UTM
        coordinate system with the origin at the southwest corner of the
        domain.
    y_coords: list[float]
        The y coordinates of the output file. Coordinates are in the UTM
        coordinate system with the origin at the southwest corner of the
        domain.
    z_coords: list[float]
        The z coordinates of the output file. Coordinates are in meters above
        the ground level.
    delimiter: str
        The delimiter used in the output file name.
    extension: str
        The file extension of the output file.
    description: str
        A description of the output file.
    units: str
        The units of the output file.
    filepaths: list[Path]
        A list of file paths for each timestep.
    times: list[int]
        A list of times corresponding to the timesteps.
    """

    def __init__(
        self,
        name: str,
        file_format: str,
        dimensions: list[str],
        grid: str,
        x_coords: list[float],
        y_coords: list[float],
        z_coords: list[float],
        delimiter: str,
        extension: str,
        description: str,
        units: str,
        times: list[int],
        filepaths: list[Path],
        index_map=None,
        crs: str = None,
    ):
        self.name = name
        self.file_format = file_format
        self.dimensions = dimensions
        self.x_coords = x_coords
        self.y_coords = y_coords
        self.z_coords = z_coords
        self.crs = crs
        self.shape = (len(z_coords), len(y_coords), len(x_coords))
        self.grid = grid
        self.delimiter = delimiter
        self.extension = extension
        self.description = description
        self.units = units
        self.times = times
        self.filepaths = filepaths
        function_mappings = {
            "gridded": _process_gridded_bin,
            "compressed": _process_compressed_bin,
        }
        self._output_function = function_mappings.get(self.file_format)
        if self._output_function is None:
            raise ValueError(f"Unknown output format: {self.file_format}")
        self._compressed_index_map = (
            index_map if (self.file_format == "compressed") else None
        )

    def __eq__(self, other):
        self_dict = self.__dict__.copy()
        other_dict = other.__dict__.copy()
        for key in self_dict.keys():
            try:
                if self_dict[key] != other_dict[key]:
                    return False
            except KeyError:
                return False
            except ValueError:
                if np.any(self_dict[key] != other_dict[key]):
                    return False
        return True

    def _select_files_based_on_timestep(
        self, timestep: int | list[int] | range | None
    ) -> list[Path]:
        """Return files selected based on timestep."""
        if timestep is None:
            return self.filepaths
        if isinstance(timestep, int):
            timestep = [timestep]
        try:
            return [self.filepaths[ts] for ts in timestep]
        except IndexError:
            raise ValueError(f"Invalid timestep: {timestep}")

    def _get_single_timestep(self, output_file: Path) -> np.ndarray:
        """Return a numpy array for the given output file."""
        return self._output_function(
            output_file, self.shape, self._compressed_index_map
        )

    def _get_multiple_timesteps(self, output_files: list[Path]) -> np.ndarray:
        """Return a numpy array for the given output files."""
        arrays = [self._get_single_timestep(of) for of in output_files]
        return arrays[0] if len(arrays) == 1 else np.concatenate(arrays, axis=0)

    def to_numpy(self, timestep: int | list[int] | range = None) -> np.ndarray:
        """
        Return a numpy array for the given output and timestep(s) with shape
        (time, nz, ny, nx). If timestep is None, then all timesteps are
        returned.

        Parameters
        ----------
        timestep: int | list[int]
            The timestep(s) to return. If None, then all timesteps are returned.

        Returns
        -------
        np.ndarray
            A 4D numpy array with shape (time, nz, ny, nx) containing the
            output data.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs", 50, 100, 100)
        >>> fire_energy = outputs.get_output("fire-energy_to_atmos")
        >>> # Get all timesteps for the fire-energy_to_atmos output
        >>> fire_energy_all = fire_energy.to_numpy()
        >>> # Get the first timestep for the fire-energy_to_atmos output
        >>> fire_energy_slice = fire_energy.to_numpy(timestep=0)
        """
        selected_files = self._select_files_based_on_timestep(timestep)
        return self._get_multiple_timesteps(selected_files)

    def to_netcdf(self, directory: str | Path, timestep: int | list[int] = None):
        """
        Write a netCDF file for the given output and timestep(s) with dimensions
        (time, nz, ny, nx). If timestep is None, then all timesteps are
        returned.

        Parameters
        ----------
        directory: str | Path
            The path to the folder where netCDF file will be written.
        timestep: int | list[int] | None
            The timestep(s) to write. If None, then all timesteps are written.

        Builds
        -------
        netCDF file to disk

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs", 50, 100, 100)
        >>> fire_energy = outputs.get_output("fire-energy_to_atmos")
        >>> out_dir = Path(path/to/output/dir)
        >>> # Get all timesteps for the fire-energy_to_atmos output
        >>> fire_energy_all = fire_energy.to_netcdf(directory = out_dir)
        >>> # Get the first timestep for the fire-energy_to_atmos output
        >>> fire_energy_slice = fire_energy.to_netcdf(directory = out_dir, timestep=0)
        """
        if isinstance(directory, str):
            directory = Path(directory)

        if not directory.exists():
            directory.mkdir(parents=True)

        if timestep is None:
            times = self.times
        elif isinstance(timestep, int):
            times = [self.times[timestep]]
        else:
            times = [self.times[i] for i in timestep]

        dataset = Dataset(directory / f"{self.name}.nc", "w", format="NETCDF4")
        dataset.title = self.name
        dataset.subtitle = self.description

        # Time dimension
        dataset.createDimension("time", None)
        dataset_time = dataset.createVariable("time", np.int64, ("time",))
        dataset_time.long_name = "Time since start of simulation"
        dataset_time.units = "s"
        dataset_time[:] = np.array(times)

        # Z dimension
        dataset.createDimension("z", None)
        dataset_z = dataset.createVariable("z", np.int64, ("z",))
        dataset_z.long_name = "Cell height: bottom of cell from ground"
        dataset_z.units = "m"
        dataset_z[:] = self.z_coords

        # Y dimension
        dataset.createDimension("y", None)
        dataset_y = dataset.createVariable("y", np.int64, ("y",))
        dataset_y.long_name = "Northing: y-coordinate of bottom edge of the cell"
        dataset_y.units = "m"
        dataset_y[:] = self.y_coords

        # X dimension
        dataset.createDimension("x", None)
        dataset_x = dataset.createVariable("x", np.int64, ("x",))
        dataset_x.long_name = "Easting: x-coordinate of left edge of the cell"
        dataset_x.units = "m"
        dataset_x[:] = self.x_coords

        # Output variable
        output = dataset.createVariable(self.name, np.float32, ("time", "z", "y", "x"))
        output.long_name = self.description
        output.units = self.units

        selected_files = self._select_files_based_on_timestep(timestep)
        output[:, :, :, :] = self._get_multiple_timesteps(selected_files)

        dataset.close()
        return

    def to_zarr(
        self,
        fpath: str | Path,
        chunk_size: dict[str, int] = None,
        **kwargs,
    ) -> zarr.Group:
        """
        Write the outputs to a zarr file.

        Parameters
        ----------
        fpath: str | Path
            The path to the folder where zarr files are written to be written.
        chunk_size: dict[str, int]
            The chunk size for the zarr array. The dictionary may contain
            keys "time", "z", "y", and "x" with integer values for the chunk
            size in each dimension. The default chunk size is (1, nz, ny, nx).
            If a key is not present, then the chunk size defaults to the size
            of the dimension.
        **kwargs
            Additional keyword arguments to pass to the zarr array creation
            function.

        Returns
        -------
        zarr.Group
            The zarr group containing the output data.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs", 50, 100, 100)
        >>> fire_energy = outputs.get_output("fire-energy_to_atmos")
        >>> # Write fire-energy_to_atmos output to zarr
        >>> zarr_group = fire_energy.to_zarr("path/to/zarr")
        >>> zarr_group["fuels-dens"]
        <zarr.core.Array '/fuels-dens' (2, 1, 200, 200) float64 read-only>
        >>> # Write multiple timesteps to each zarr chunk to improve I/O
        >>> zarr_group = fire_energy.to_zarr("path/to/zarr", chunk_size={"time": 100})
        >>> zarr_group["fuels-dens"].chunks
        (100, 1, 200, 200)
        """
        if isinstance(fpath, str):
            fpath = Path(fpath)

        if not fpath.exists():
            fpath.mkdir(parents=True)

        # Build the zarr file with directory storage
        zarr_path = fpath / (self.name + ".zarr")
        zroot = zarr.open(str(zarr_path), mode="w")
        zroot.attrs["crs"] = self.crs

        # Build the output data array. By default, the array has shape
        # (time, nz, ny, nx) and chunks of size (1, nz, ny, nx).
        shape = (len(self.times), *self.shape)
        if chunk_size is None:
            chunk_size = {}
        chunks = [
            chunk_size.get("time", 1),
            chunk_size.get("z", self.shape[0]),
            chunk_size.get("y", self.shape[1]),
            chunk_size.get("x", self.shape[2]),
        ]
        output_array = zroot.create_dataset(
            self.name, shape=shape, chunks=chunks, dtype=float, **kwargs
        )
        output_array.attrs["_ARRAY_DIMENSIONS"] = ["time", "z", "y", "x"]
        output_array.attrs["long_name"] = self.description
        output_array.attrs["units"] = self.units

        # Write each timestep to the output array
        for time_step in range(len(self.times)):
            data = self.to_numpy(time_step)
            output_array[time_step, ...] = data[0, ...]

        # Time dimension
        time_array = zroot.create_dataset("time", shape=len(self.times), dtype=int)
        time_array.attrs["long_name"] = "Time since start of simulation"
        time_array.attrs["units"] = "s"
        time_array[...] = self.times

        # Z dimension
        z_array = zroot.create_dataset("z", shape=self.shape[0], dtype=float)
        z_array.attrs["long_name"] = "Cell height: bottom of cell from ground"
        z_array.attrs["units"] = "m"
        z_array[...] = self.z_coords

        # Y dimension
        y_array = zroot.create_dataset("y", shape=self.shape[1], dtype=float)
        y_array.attrs[
            "long_name"
        ] = "Northing: cell distance north from southern edge of domain"
        y_array.attrs["units"] = "m"
        y_array[...] = self.y_coords

        # X dimension
        x_array = zroot.create_dataset("x", shape=self.shape[2], dtype=float)
        x_array.attrs[
            "long_name"
        ] = "Easting: cell distance east from western edge of domain"
        x_array.attrs["units"] = "m"
        x_array[...] = self.x_coords

        zarr.consolidate_metadata(zarr_path)

        return zroot


class SimulationOutputs:
    """
    A class responsible for managing and processing simulation outputs,
    including validation, extraction, and organization of data from output
    files. This class facilitates the retrieval of data in various formats
    and can return a numpy array for a specific output, or write the data to a
    zarr file.

    Attributes
    ----------
    output_directory: Path
        The path to the directory containing the simulation outputs. QUIC-Fire
        defaults this to the "Output" directory in the same directory as the
        simulation input files, but that does not have to be the case for
        quicfire-tools. The directory must contain the "fire_indexes.bin" and
        "grid.bin" files.
    fire_nz: int
        The number of vertical cells in the fire grid.
    ny: int
        The number of cells in the y-direction.
    nx: int
        The number of cells in the x-direction.
    dy: float
        The grid spacing in the y-direction (m).
    dx: float
        The grid spacing in the x-direction (m).
    utm_x: float
        UTM-x coordinates of the south-west corner of domain [m]. Default is 0.
    utm_y: float
        UTM-y coordinates of the south-west corner of domain [m]. Default is 0.
    utm_zone: int
        UTM zone of the domain. Default is None. If None, the coordinates are
        considered not to be georeferenced.
    crs: str
        EPSG code for the coordinate reference system. If the UTM zone is
        provided, the crs is set to the UTM zone. If the UTM zone is not
        provided, the crs is set to None.
    """

    def __init__(
        self,
        output_directory: Path | str,
        fire_nz: int,
        ny: int,
        nx: int,
        dy: float,
        dx: float,
        utm_x: float = 0,
        utm_y: float = 0,
        utm_zone: int = 1,
    ):
        if isinstance(output_directory, str):
            output_directory = Path(output_directory)
        self._validate_output_dir(output_directory)

        # Assign attributes
        self.output_directory = output_directory
        self.ny = ny
        self.nx = nx
        self.dy = dy
        self.dx = dx
        self.fire_nz = fire_nz
        self.utm_x = utm_x
        self.utm_y = utm_y
        self.utm_zone = utm_zone

        # TODO: fire_nz can be optional
        # TODO: Throw warning if fire_nz not provided and qf_wind outputs are present
        # TODO: raise error if fire_nz is not provided and qf_qind is output

        # Compute the x and y coordinate grids
        self.x_coords = np.linspace(utm_x, utm_x + dx * (nx - 1), nx)
        self.y_coords = np.linspace(utm_y, utm_y + dy * (ny - 1), ny)

        # Set the crs if the UTM zone is provided. The if statement here is
        # looking for the default QUIC-Fire values of 1, 0, 0. If the UTM
        # zone is not 1 or the x and y coordinates are not 0, then the crs
        # is set to the UTM zone.
        if utm_zone != 1 or (utm_x != 0 and utm_y != 0):
            crs = proj.CRS.from_proj4(f"+proj=utm +zone={utm_zone} +ellps=WGS84")
            self.crs = f"EPSG:{crs.to_epsg()}"
        else:
            self.crs = None

        # Get grid information from grid.bin and fire_indexes.bin
        self._fire_indexes = _process_fire_indexes(
            output_directory / "fire_indexes.bin"
        )
        (
            self.quic_nz,
            self.quic_z_coords,
            self.en2atmos_nz,
            self.en2atmos_z_coords,
        ) = _process_grid_info(output_directory / "grid.bin", ny, nx)
        self.quic_z_coords = self.quic_z_coords[: self.quic_nz]
        self.en2atmos_z_coords = self.en2atmos_z_coords[: self.en2atmos_nz]
        self.fire_z_coords = self.en2atmos_z_coords[:fire_nz]
        self.quic_dz = _get_resolution_from_coords(self.quic_z_coords)
        self.fire_dz = _get_resolution_from_coords(self.en2atmos_z_coords)

        # Build a list of present output files and their times
        self.outputs = {}
        self._build_output_files_map()
        self._output_names = list(self.outputs.keys())

    def __iter__(self):
        """Return an iterator for the object."""
        self._index = 0
        return self

    def __next__(self):
        """Return the next item in the iterator."""
        if self._index < len(self._output_names):
            output = self.outputs[self._output_names[self._index]]
            self._index += 1
            return output
        else:
            raise StopIteration

    def __eq__(self, other):
        self_dict = self.__dict__.copy()
        other_dict = other.__dict__.copy()
        self_dict.pop("outputs")
        other_dict.pop("outputs")
        for key in self_dict.keys():
            self_value = self_dict[key]
            other_value = other_dict[key]
            try:
                if self_value != other_value:
                    return False
            except KeyError:
                return False
            except ValueError:
                if np.any(self_value != other_value):
                    return False
        for key in self.outputs.keys():
            try:
                if self.outputs[key] != other.outputs[key]:
                    return False
            except KeyError:
                return False
        return True

    @classmethod
    def from_simulation_inputs(
        cls, output_directory: Path | str, simulation_inputs: SimulationInputs
    ) -> SimulationOutputs:
        """
        Create a SimulationOutputs object from a path to a QUIC-Fire "Output"
        directory and a SimulationInputs object.

        Parameters
        ----------
        output_directory: Path | str
            The path to the directory containing the simulation outputs. The
            directory must contain the "fire_indexes.bin" and "grid.bin" files.
            This is typically the "Output" directory in the same directory as
            the simulation input files, but that does not have to be the case
            for quicfire-tools.
        simulation_inputs: SimulationInputs
            The SimulationInputs object containing the simulation input data.

        Returns
        -------
        SimulationOutputs
            A SimulationOutputs object.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> inputs = qft.SimulationInputs.from_directory("path/to/inputs")
        >>> outputs = qft.SimulationOutputs.from_simulation_inputs("path/to/outputs", inputs)
        """
        return cls(
            output_directory,
            simulation_inputs.quic_fire.nz,
            simulation_inputs.qu_simparams.ny,
            simulation_inputs.qu_simparams.nx,
            simulation_inputs.qu_simparams.dy,
            simulation_inputs.qu_simparams.dx,
            simulation_inputs.qu_simparams.utm_x,
            simulation_inputs.qu_simparams.utm_y,
            simulation_inputs.qu_simparams.utm_zone_number,
        )

    @staticmethod
    def _validate_output_dir(outputs_directory):
        """
        Validate the outputs directory by checking for the existence of the
        Output directory and the required files.
        """
        # Check if the outputs directory exists and is a directory
        if not outputs_directory.exists():
            raise FileNotFoundError(
                f"The directory {outputs_directory} does not exist."
            )
        elif not outputs_directory.is_dir():
            raise NotADirectoryError(
                f"The path {outputs_directory} is not a directory."
            )

        # Check for required files
        required_files = ["fire_indexes.bin", "grid.bin"]
        for f in required_files:
            if not (outputs_directory / f).exists():
                raise FileNotFoundError(f"Required file {f} not in outputs directory")

    def _build_output_files_map(self):
        """
        Build a key-value map of output files, where the key is the name of
        the output file and the value is an instantiated OutputFile object.
        """
        for key, attributes in OUTPUTS_MAP.items():
            output_files_list = self._get_list_output_paths(
                key, attributes["extension"]
            )

            # Check if any output files of the given type exist in the directory
            if output_files_list:
                z_coords = self._get_z_coords(attributes)
                times = []
                for file in output_files_list:
                    time = self._get_output_file_time(key, file)
                    times.append(time)

                self.outputs[key] = OutputFile(
                    name=key,
                    file_format=attributes["file_format"],
                    dimensions=attributes["dimensions"],
                    grid=attributes["grid"],
                    x_coords=self.x_coords,
                    y_coords=self.y_coords,
                    z_coords=z_coords,
                    delimiter=attributes["delimiter"],
                    extension=attributes["extension"],
                    description=attributes["description"],
                    units=attributes["units"],
                    times=times,
                    filepaths=output_files_list,
                    index_map=self._fire_indexes,
                    crs=self.crs,
                )

    def _get_list_output_paths(self, name, ext) -> list[Path]:
        """
        Get a sorted list of output files in the Output/ directory for the
        given output name.
        """
        paths = list(self.output_directory.glob(f"{name}*{ext}"))
        paths.sort()
        return paths

    def _get_z_coords(self, attrs) -> tuple:
        number_dimensions = len(attrs["dimensions"])
        grid = attrs["grid"]

        if number_dimensions == 2:
            return [0.0]
        elif number_dimensions == 3 and grid == "fire":
            return self.fire_z_coords
        elif number_dimensions == 3 and grid == "en2atmos":
            return self.en2atmos_z_coords
        elif number_dimensions == 3 and grid == "quic":
            return self.quic_z_coords
        else:
            raise ValueError(
                f"Invalid number of dimensions ({number_dimensions}) for {grid} grid"
            )

    @staticmethod
    def _get_output_file_time(output, fpath) -> int:
        """Get the time of the output file."""
        delimiter = OUTPUTS_MAP[output]["delimiter"]
        extension = OUTPUTS_MAP[output]["extension"]

        try:
            if delimiter == "":
                # Split on first numeric component
                name_parts = re.split(r"(\d+)", fpath.name)
                file_time = name_parts[1]
            else:
                file_time = fpath.name.split(delimiter)[-1].split(extension)[0]

            return int(file_time)

        except ValueError:  # No numeric component
            return 0

    def _validate_output(self, output: str) -> OutputFile:
        """Validate output."""
        try:
            output = self.outputs[output]
        except KeyError:
            raise ValueError(
                f"{output} is not a valid output. Valid "
                f"outputs are: {self.list_outputs()}"
            )
        return output

    def list_outputs(self) -> list[str]:
        """
        Return a list of available output names.

        Returns
        -------
        list[str]
            A list of available output names.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs")
        >>> outputs.list_outputs()
        ['fire-energy_to_atmos', 'fuels-dens', 'groundfuelheight', 'mburnt_integ']
        """
        return self._output_names

    def get_output(self, key) -> OutputFile:
        """
        Return the OutputFile object for the given output name.

        Parameters
        ----------
        key: str
            The name of the output to return.

        Returns
        -------
        OutputFile
            The OutputFile object for the given output name.

        Raises
        ------
        ValueError
            If the output name is not valid.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs")
        >>> fire_energy = outputs.get_output("fire-energy_to_atmos")
        >>> fire_energy
        <quicfire_tools.outputs.OutputFile object at 0x7f8b1c2b6d90>
        """
        return self._validate_output(key)

    def to_numpy(
        self, key: str, timestep: None | int | list[int] | range = None
    ) -> np.ndarray:
        """
        Returns a 4D numpy array for the given output and timestep(s) with shape
        (time, nz, ny, nx). If timestep is None, then all timesteps are
        returned.

        Parameters
        ----------
        key: str
            The name of the output to return.
        timestep: int | list[int] | None
            The timestep(s) to return. If None, then all timesteps are returned.

        Returns
        -------
        np.ndarray
            A 4D numpy array with shape (time, nz, ny, nx) containing the
            output data.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs", 50, 100, 100)
        >>> output_name = "fire-energy_to_atmos"
        >>>
        >>> # Get all timesteps for the fire-energy_to_atmos output
        >>> fire_energy_all = outputs.to_numpy(output_name)
        >>>
        >>> # Get the first timestep for the fire-energy_to_atmos output
        >>> fire_energy_first_time_step = outputs.to_numpy(output_name, timestep=0)
        """
        output = self._validate_output(key)
        return output.to_numpy(timestep)

    def to_dask(self, key: str) -> da.Array:
        """
        Returns a dask array for the given output with shape (time, nz, ny, nx).

        Parameters
        ----------
        key: str
            The name of the output to return.

        Returns
        -------
        da.Array
            A dask array with shape (time, nz, ny, nx) containing the output
            data with chunks of size (1, nz, ny, nx). The dask array is
            lazily evaluated, so users must call `.compute()` to retrieve the
            data.

        Examples
        --------
        >>> import quicfire_tools as qft
        >>> outputs = qft.SimulationOutputs("path/to/outputs", 50, 100, 100)
        >>> fire_energy_da = outputs.to_dask("fire-energy_to_atmos")
        >>> fire_energy_np = fire_energy_da.compute()  # Retrieve the data
        """
        output = self._validate_output(key)

        # Create a dask array for the output file
        shape = (len(output.times), *output.shape)
        chunks = [1 if i == 0 else shape[i] for i in range(len(shape))]
        dask_array = da.zeros(shape, dtype=float, chunks=chunks)

        # Write each timestep to the dask array
        for time_step in range(len(output.times)):
            data = self.to_numpy(output.name, time_step)
            dask_array[time_step, ...] = data

        return dask_array


def _get_resolution_from_coords(coords: list[float]) -> float | list[float]:
    """
    Get the resolution from a list of coordinates. If the resolution is
    constant, then a single float is returned. If the resolution is not
    constant, then a list of floats is returned.
    """
    resolution = np.diff(coords)
    if np.all(resolution == resolution[0]):
        return resolution[0]
    else:
        return resolution


def _process_grid_info(path_to_grid_bin: str | Path, ny: int, nx: int):
    """
    Reads grid information from the grid.bin file and populates attributes
    for fire and quic grid sizes, as well as dx, dy, and dz.
    """
    if isinstance(path_to_grid_bin, str):
        path_to_grid_bin = Path(path_to_grid_bin)

    with open(path_to_grid_bin, "rb") as fid:
        # Read the number of bytes in the vertical QUIC grid
        num_bytes_quic_grid = np.fromfile(fid, dtype=np.int32, count=1)[0]

        # Get the number of cells in the vertical direction
        # num_cells * 32 bits = num_bytes * (8 bits / byte)
        quic_nz = int(num_bytes_quic_grid / 4) - 2

        # QUIC-URB grid bottom values
        quic_grid_bottom = np.fromfile(fid, dtype=np.float32, count=quic_nz + 2)

        np.fromfile(fid, dtype=np.int32, count=2)  # Header
        np.fromfile(fid, dtype=np.float32, count=quic_nz + 2)  # quic grid mid values
        np.fromfile(fid, dtype=np.int32, count=1)  # Header

        # Try and read a topo or a fire header
        test_header = np.fromfile(fid, dtype=np.int32, count=1)
        if len(test_header) == 0:
            raise ValueError("No fire or topo grids found in grid.bin file")

        # Check if topo grid information is present
        if test_header[0] == num_bytes_quic_grid:
            np.fromfile(
                fid, dtype=np.float32, count=quic_nz + 2
            )  # Read sigma_bottom (same as quic_grid_bottom)
            np.fromfile(fid, dtype=np.int32, count=2)  # Header
            np.fromfile(
                fid, dtype=np.float32, count=quic_nz + 2
            )  # Read sigma_mid (same as quic_grid_mid)
            np.fromfile(fid, dtype=np.int32, count=2)  # Header
            np.fromfile(
                fid, dtype=np.float32, count=ny * nx * (quic_nz + 2)
            )  # quic grid bottom terrain following
            np.fromfile(fid, dtype=np.int32, count=2)  # Header
            np.fromfile(
                fid, dtype=np.float32, count=ny * nx * (quic_nz + 2)
            )  # quic grid mid terrain following
            np.fromfile(fid, dtype=np.int32, count=2)  # Header
            np.fromfile(
                fid, dtype=np.float32, count=ny * nx
            )  # quic grid volume correction
            np.fromfile(fid, dtype=np.int32, count=1)  # Header
            _ = np.fromfile(fid, dtype=np.int32, count=1)[0]  # fire header
        else:
            _ = test_header[0]  # fire header

        num_eng_to_atmos_cells = np.fromfile(fid, dtype=np.int32, count=1)[0]
        np.fromfile(fid, dtype=np.int32, count=2)  # Header
        eng_to_amos_grid_bottom = np.fromfile(
            fid, dtype=np.float32, count=num_eng_to_atmos_cells + 1
        )

        np.fromfile(fid, dtype=np.int32, count=1)  # ending header
        np.fromfile(fid, dtype=np.int32, count=4)  # Header

        return (
            quic_nz,
            quic_grid_bottom[:],
            num_eng_to_atmos_cells,
            eng_to_amos_grid_bottom[:],
        )


def _process_fire_indexes(path_to_fire_indexes_bin: str | Path) -> ndarray:
    """
    Process the `fire_indexes.bin` file to extract the necessary
    information to rebuild the 3D fields of fire and fuel related variables.
    Function returns a 2D ndarray of shape (num_cells, 3) representing the
    i, j, k indexes of each cell that contains data in the sparse (compressed)
    format.
    """
    if isinstance(path_to_fire_indexes_bin, str):
        path_to_fire_indexes_bin = Path(path_to_fire_indexes_bin)

    with open(path_to_fire_indexes_bin, "rb") as fid:
        np.fromfile(fid, dtype=np.int32, count=1)  # Header
        num_cells = np.fromfile(fid, dtype=np.int32, count=1)[0]
        np.fromfile(fid, dtype=np.int32, count=2)  # Header
        np.fromfile(fid, dtype=np.int32, count=1)  # max k index
        np.fromfile(fid, dtype=np.int32, count=2)  # Header
        np.fromfile(fid, dtype=np.int32, count=num_cells)  # unique cell IDs
        np.fromfile(fid, dtype=np.int32, count=2)  # Header

        # Initialize an array to hold the indices
        ijk = np.zeros((num_cells, 3), dtype=np.int32)

        # Loop over each dimension's indices (Like a sparse meshgrid)
        for i in range(0, 3):
            ijk[:, i] = np.fromfile(fid, dtype=np.int32, count=num_cells)

        # Convert to indices (numbered from 0)
        ijk -= 1

        return ijk


def _process_compressed_bin(filename, dim_zyx, *args) -> ndarray:
    """
    Converts the contents of a sparse .bin file to a dense array with shape
    (1, nz, ny, nx).

    The mapping from sparse to dense is guided by an index map provided as an
    argument. The index map must match the sparse data in the .bin file,
    defining the locations of nonzero values within the dense array.

    Parameters
    ----------
    filename : str | Path
        The name or path of the sparse .bin file to be processed.
    dim_zyx : tuple of int
        The number of z, y, x cells to store in the 3D array, given as a tuple
        of three integers.
    *args : tuple
        Additional arguments, specifically expecting the index map for mapping
        sparse indices to dense. The index map should be a 2D array with shape
        (N, 3), where N is the number of nonzero elements in the sparse data.

    Returns
    -------
    ndarray
        A dense 4D array of shape (1, nz, ny, nx) containing the sparse data.

    Raises
    ------
    IndexError
        If the index map (fire_indexes) is not provided in the arguments.
    ValueError
        If the index map (fire_indexes) is None.
    """
    try:
        fire_indexes = args[0]
    except IndexError:
        raise ValueError(
            "fire_indexes must be provided when processing " "compressed .bin files."
        )
    if fire_indexes is None:
        raise ValueError(
            "fire_indexes must be provided when processing " "compressed .bin files."
        )

    # Initialize the 3D array
    full_array = np.zeros([1, *dim_zyx], dtype=np.float32)
    array_3d = full_array[0, :, :, :]

    with open(filename, "rb") as fid:
        # Read header
        np.fromfile(fid, dtype=np.int32, count=1)

        # Read in the sparse values
        sparse_values = np.fromfile(fid, dtype=np.float32, count=fire_indexes.shape[0])

        # Map indices of the sparse data to the indices of the dense array
        indices = (
            fire_indexes[:, 2],  # k indices (z)
            fire_indexes[:, 1],  # j indices (y)
            fire_indexes[:, 0],  # i indices (x)
        )

        # Update the full array with the sparse indices
        array_3d[indices] = sparse_values

    full_array[0, ...] = array_3d
    return full_array


def _process_gridded_bin(filename, dims_zyx, *args) -> ndarray:
    """
    Converts the data stored in a gridded .bin file to an np array. The
    function takes in a 3D array of dimensions (nz, ny, nx) and returns an
    array of shape (1, nz, ny, nx).

    Parameters
    ----------
    filename : str | Path
        File path to the .bin file that contains the gridded data.
    dims_zyx : tuple
        A tuple of length 3 containing the number of z, y, and x cells in the
        3D array.
    *args : list
        Additional arguments that might be passed to the underlying slice
        processing function (_process_gridded_bin_slice).

    Returns
    -------
    ndarray
        A 4D array of shape (1, nz, ny, nx) containing the gridded data.
    """
    # Initialize the array
    array = np.zeros([1, *dims_zyx], dtype=np.float32)

    nz, ny, nx = dims_zyx
    with open(filename, "rb") as f:
        # Read header
        _ = np.fromfile(f, dtype=np.float32, count=1)

        # Read in gridded data
        for k in range(nz):
            array[0, k, :, :] = _process_gridded_bin_slice(f, ny, nx)
    return array.astype(float)


def _process_gridded_bin_slice(fid, ny, nx) -> ndarray:
    """
    Process a 2D slice from a gridded .bin file.

    This method reads a 2D slice of size (ny, nx) from the provided file
    object, and returns it as a np.float32 array.
    """
    # Read in gridded data
    plane = np.fromfile(fid, dtype=np.float32, count=ny * nx)
    return np.reshape(plane, (ny, nx))



"""
Utility functions for quicfire-tools.
"""

from __future__ import annotations

import numpy as np
from pathlib import Path
from numpy import ndarray
from scipy.io import FortranFile


def compute_parabolic_stretched_grid(
    dz_surf: float, n_surf_cell: int, n_tot_cells: int, domain_height: float
) -> np.ndarray:
    """
    Generate a vertical grid spacing array with parabolic stretching.

    Parameters
    ----------
    dz_surf : float
        Vertical cell spacing at the surface, typically about 1 m.
    n_surf_cell : int
        Number of cells with constant vertical cell spacing (dz_surf) at the
        surface.
    n_tot_cells : int
        Total number of vertical cells in the grid. Typically ranges between
        20 and 30.
    domain_height : float
        Total height of the grid. For flat terrain without smoke transport
        concerns, it is usually around 100 m. For terrain with features, it
        should be approximately 3 times the height of the tallest feature.

    Returns
    -------
    np.ndarray
        Array representing the vertical grid spacing (dz) values for the
        simulation.

    Notes
    -----
    The function computes a parabolic stretching of the vertical grid spacings.
    The grid consists of a specified number of cells with uniform spacing at
    the surface, followed by cells with spacings determined by a parabolic
    formula until the total domain height is achieved.
    """
    dzmax_high = domain_height - dz_surf * n_surf_cell
    dzmax_low = 0

    dz = np.ones(n_tot_cells) * dz_surf

    while True:
        dzmax = 0.5 * (dzmax_low + dzmax_high)

        c1 = (dzmax - dz_surf) / ((n_tot_cells - n_surf_cell) ** 2)
        c2 = -2.0 * c1 * n_surf_cell
        c3 = dz_surf + c1 * n_surf_cell**2

        # Apply parabolic formula for cells beyond the surface cells
        kreal = np.arange(n_surf_cell, n_tot_cells, dtype=float)
        dz[n_surf_cell:] = (c1 * kreal**2) + (c2 * kreal) + c3

        zmax_temp = np.sum(dz)

        if abs(domain_height - zmax_temp) < 0.001:
            break
        elif zmax_temp > domain_height:
            dzmax_high = dzmax
        else:
            dzmax_low = dzmax

    dz[-1] += domain_height - np.sum(dz)

    return dz


def read_dat_file(
    filename: Path | str, nz: int, ny: int, nx: int, order: str = "C"
) -> ndarray:
    """
    Read in a .dat file as a numpy array.

    Parameters
    ----------
    filename : Path or str
        The path to the .dat file to read.
    nz : int
        The number of cells in the z-direction.
    ny : int
        The number of cells in the y-direction.
    nx : int
        The number of cells in the x-direction.
    order : str, optional
        The order of the array. Default is "C".

    Returns
    -------
    ndarray
        A 3D numpy array representing the data in the .dat file. The array
        has dimensions (nz, ny, nx).
    """
    if isinstance(filename, str):
        filename = Path(filename)

    with open(filename, "rb") as fin:
        arr = (
            FortranFile(fin)
            .read_reals(dtype="float32")
            .reshape((nz, ny, nx), order=order)
        )

    return arr
